package jrf.listener;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.concurrent.TimeUnit;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.FileOutputStream;

import jrf.hbset.util.ACHSet;
import jrf.hbset.util.AbstractArrayImpl;
import jrf.hbset.util.HBDataRaceException;
import jrf.hbset.util.PathElementList;
import jrf.eliminator.DataRaceEliminator;
import jrf.hbset.util.Key;
import jrf.hbset.util.MyMap;
import jrf.hbset.util.MySet;
import jrf.search.HBDFSearch;
import jrf.search.HBBFSearch;
import jrf.search.HBHeuristicSearch;

import gov.nasa.jpf.PropertyListenerAdapter;
import gov.nasa.jpf.jvm.ElementInfo;
import gov.nasa.jpf.jvm.FieldInfo;
import gov.nasa.jpf.jvm.ClassInfo;
import gov.nasa.jpf.jvm.MethodInfo;
import gov.nasa.jpf.jvm.JVM;
import gov.nasa.jpf.jvm.Path;
import gov.nasa.jpf.jvm.StackFrame;
import gov.nasa.jpf.jvm.StateSet;
import gov.nasa.jpf.jvm.Step;
import gov.nasa.jpf.jvm.ThreadInfo;
import gov.nasa.jpf.jvm.bytecode.FieldInstruction;
import gov.nasa.jpf.jvm.bytecode.LockInstruction;
import gov.nasa.jpf.jvm.bytecode.InstanceFieldInstruction;
import gov.nasa.jpf.jvm.bytecode.Instruction;
import gov.nasa.jpf.jvm.bytecode.InvokeInstruction;
import gov.nasa.jpf.jvm.bytecode.GETFIELD;
import gov.nasa.jpf.jvm.bytecode.GETSTATIC;
import gov.nasa.jpf.jvm.bytecode.NEW;
import gov.nasa.jpf.jvm.bytecode.NEWARRAY;
import gov.nasa.jpf.jvm.bytecode.ANEWARRAY;
import gov.nasa.jpf.jvm.bytecode.MULTIANEWARRAY;
import gov.nasa.jpf.jvm.bytecode.ReturnInstruction;
import gov.nasa.jpf.jvm.bytecode.StaticFieldInstruction;
import gov.nasa.jpf.jvm.bytecode.ArrayLoadInstruction;
import gov.nasa.jpf.jvm.bytecode.ArrayStoreInstruction;
import gov.nasa.jpf.search.Search;
import gov.nasa.jpf.report.Publisher;
import gov.nasa.jpf.report.ConsolePublisher;
import gov.nasa.jpf.report.Statistics;
import gov.nasa.jpf.Config;
import gov.nasa.jpf.JPF;

import jrf.hbset.util.Log;
import jrf.hbset.ListenerHBmanagement;
import static jrf.hbset.util.Utility.getThreadObject;
import static jrf.hbset.util.AbstractArrayImpl.*;


/*
 * AssertionalReasoningListener is the starting point of HB implementation
 * each listener will be called by gov.nasa.jpf.JVM or gov.nasa.jpf.ThreadInfo
 */
public class JRFListener extends PropertyListenerAdapter {
	// to indicate search is started
	// note that many information is not consistent before search start 
	// since they might not be initialized yet by JPF
	private static boolean Initialized=false;

	// store all initially created objects and loaded classes before search start
	// including main thread, main thread group, and Thread.java etc.
	// HB for these will be explicitly created at search start  

	private static MySet<ElementInfo> delayed_object_creation = new MySet<ElementInfo>();
	private static MySet<ClassInfo> delayed_class_creation = new MySet<ClassInfo>();


	/***********************************************/
	/* information to store at state store/restore */
	/***********************************************/

	// this method check if the thread is inside static initializer
	// objects created in static initializer will be handled differently at ListenerHBManagement
	public static boolean isWithinStaticInit(ThreadInfo ti)
	{
		return ti.isClinitOnStack();
	}

	// store the elements used in each instruction
	//PPP private static ArrayList<Key> pathElements = new ArrayList<Key>();

	private static PathElementList pathElements = new PathElementList();

	// we store the index of pathElements vector for each state
	// StateBacktrack will recover previous state pathElements index
	private static MyMap<Integer, Integer> statePathIndex = new MyMap<Integer, Integer>();

	// StateStore will store copy of these information, (State) 
	// and these will be restored to stored value at StateRestore 
	static class State {
		final PathElementList pathElements;
		final MyMap<Integer, Integer> statePathIndex;
		final Object hbData;

		State(PathElementList _pathElements, 
			  MyMap<Integer, Integer> _statePathIndex,
			  Object _hbData)
		{
			this.pathElements = _pathElements;
			this.statePathIndex = _statePathIndex;
			this.hbData = _hbData;
		}
	}

	// adds states at StateStore
	// removes states at StateRestore
	private static MyMap<Integer, ArrayList<State>> store = new MyMap<Integer ,ArrayList<State>>();

	/***********************************************/
	/* information to handle found races		   */
	/***********************************************/
	static final boolean multiraces = JVM.getVM().getConfig().getBoolean("jrf.multiraces");
	static final int thresholdraces = JVM.getVM().getConfig().getInt("jrf.multiraces.threshold"); // 0 means no limit
	static final boolean showtrace = JVM.getVM().getConfig().getBoolean("jrf.race.showtrace"); 
	
	// store found race for later analysis
	private static class Race {
		HBDataRaceException raceException; // save race location
		String snapshot; // snapshot cannot be obtained later, so it should be saved at the time of race
		Path path; // saves counterexample path
		int stepCount; // save the stepCounter at the time of a race, this is because the clone() gives shallow copy and path is changing
		Instruction manifestInstruction; // save raceManifestInstruction
		PathElementList pathElements; // save pathElements
		//String counterexampleAnalysis; // this is for testing purpose only
		MySet<Key> agents; // saves current h information to verify that the acquiring history applicability
		int threadIndex;
	}
	
	// stores races found during state search
	private static ArrayList<Race> races = new ArrayList<Race>();
	
	// summarizes the advices during analysis
	// mapping each advice to its rank that represents the number of times the suggestion made
	private static MyMap<String, Integer> advices = new MyMap<String, Integer>();

	
	// initialize Publisher
	public JRFListener(Config config, JPF jpf)
	{
		jpf.addPublisherExtension(ConsolePublisher.class, this);
	}
	
	
	/*** PublisherExtension ****************************************************/
	 /*
	 public void publishTransition (Publisher publisher) {}
	 public void publishStart (Publisher publisher) {}
	 public  void publishPropertyViolation (Publisher publisher) {}
	 public void publishConstraintHit (Publisher publisher) {}
	 */
	 public void publishFinished (Publisher publisher)
	 { 
		 publishResultToFile(publishRace(publisher.getOut()));
	 }
	 
	 String resultFile = "result/RESULT";
	 
	 private void publishResultToFile(JRFresult result)
	 {
		 try {
			 PrintWriter pw = new PrintWriter(new FileOutputStream(resultFile, true));
			 
			 Config config = JVM.getVM().getConfig();
			 Statistics stat = JVM.getVM().getJPF().getReporter().getStatistics();
			 Search search = JVM.getVM().getJPF().getSearch();
			 
			 String msg = config.getTarget();
			 msg += " , "+((search instanceof HBDFSearch)?
					 		((config.getInstance("vm.storage.class", StateSet.class)==null)?"STATELESS":"DFS"):
					 	  (search instanceof HBHeuristicSearch)?"HEURISTIC":
				 		  (search instanceof HBBFSearch)?"BFS":"?");
			 msg += " , "+stat.newStates;
			 msg += " , "+result.hstates;
			 msg += " , "+result.max_length;
			 msg += " , "+result.race_time+"s";
			 msg += result.analysis_time==null?"":" & "+result.analysis_time+"s";
			 msg += " , "+(stat.maxUsed >>20) + "MB";
			 msg += " , "+result.hsize;
			 msg += " , "+result.race_location; 
System.err.println("JRFresult = "+msg+"\n\n");			 
			 pw.println(msg);
			 pw.flush();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	 }
	 
	 class JRFresult {
		 String race_time;
		 String analysis_time;
		 int  	max_length;
		 int 	hstates;
		 String race_location = new String();
		 String hsize;
	 }
	 
	 private JRFresult publishRace(PrintWriter pw)
	 {
		 MyMap<String, MyMap<String, String>> raceSet = new MyMap<String, MyMap<String, String>>();
		 MyMap<String, MySet<String>> raceSuggestion = new MyMap<String, MySet<String>>();
		 //int counterPathLength=0;
		 JRFresult result = new JRFresult();
		 
		 long analysisTime = System.nanoTime();
System.err.println("\n@@@ CHECK:"+races.size()+" races are found.");

		 for ( Race r : races )
		 {
			 result.race_location += "("+r.raceException.getMemoryObject().toName()+")"
			 						+r.raceException.getSourceLine();
			 // find list for the race type
			 MyMap<String, String> msgList;
			 if ( !raceSet.containsKey(r.raceException.toString()) ) 
			 {
				 msgList = new MyMap<String, String>();
				 raceSet.put(r.raceException.toString(), msgList);
			 }
			 else 
				 msgList = raceSet.get(r.raceException.toString());
			 
			 // find suggestion set for the race type
			 MySet<String> suggestions;
			 if ( !raceSuggestion.containsKey(r.raceException.toString()) )
			 {
				 suggestions = new MySet<String>();
				 raceSuggestion.put(r.raceException.toString(), suggestions);
			 }
			 else
				 suggestions = raceSuggestion.get(r.raceException.toString());
			 
			 
			 String suggestionString = new String();

			 // print counterexample analysis
			 // clone of path is shallow copy, so it doesn't have the path information at the time of a race
			 // pathElements padded for those additional path steps
			 if ( DataRaceEliminator.choice )
			 {
				 int additional = getTotalNumInstructions(r.path)-r.stepCount;
				 if ( additional != 0 )
				 {
					 assert(r.pathElements.size()==r.stepCount);
					 PathElementList patch = r.pathElements.clone();
					 for ( int i=0 ; i < additional ; ++i )
						 patch.add(null);
					 r.pathElements = patch;
				 }
				 assert(r.pathElements.size()==getTotalNumInstructions(r.path));
			 }
			 
			 if ( DataRaceEliminator.choice )
			 {
				 ArrayList<String> cAnalysis = new DataRaceEliminator().analyzeCounterExample(r.path,
					 													getTotalNumInstructions(r.path)-r.stepCount,
					 													r.threadIndex,
																		r.manifestInstruction, 
																		r.pathElements, 
																		r.raceException.getMemoryObject());
				 suggestionString += "\n______________________________________________________ analyze counter example \n";
				 int i=0;
				 for ( String s : cAnalysis )
				 {
					 suggestionString += s+"\n";
					 if ( !s.startsWith("data race source statement") 
							 && !s.startsWith("data race manifest statement")
							 && !s.startsWith("Unsafe publication of object") ) 
						 suggestions.add(s); 
				 }
			 }
			 
			 // print acquiring history analysis
			 if (  ACHSet.choice )
			 {
				 ArrayList<String> aHistory = (ACHSet.advice(r.raceException.getMemoryObject(), 
						 				  r.raceException.getThreadObject(), 
						 				  r.agents));
				 suggestionString += "\n______________________________________________________ advice from acquiring history\n";
				 for ( String s : aHistory )
				 {
					 suggestionString += s;
					 suggestions.add(s);
				 }
			 }
			 
			 // add to analysis			 
			 if ( showtrace )
			 {
				 if ( !msgList.containsKey(suggestionString) ) 
				 {
					 // get counterexample path and snapshot
					 String trace = new String("______________________________________________________ trace #1\n");
					 trace += (DataRaceEliminator.getCounterexamplePath(JVM.getVM(), r.path, r.stepCount));
					 trace += ("______________________________________________________ snapshot #1\n");
					 trace += r.snapshot;
					 msgList.put(suggestionString, trace);
				 }
				 else
				 {
					 String existing = msgList.remove(suggestionString);
					 int traceNum = existing.split("______________________________________________________ trace #").length; 
					 // get counterexample path and snapshot
					 String trace = new String("______________________________________________________ trace #"+traceNum+"\n");
					 trace += (DataRaceEliminator.getCounterexamplePath(JVM.getVM(), r.path, r.stepCount));
					 trace += ("______________________________________________________ snapshot #"+traceNum+"\n");
					 trace += r.snapshot;
					 msgList.put(suggestionString, existing+trace);
				 }
		 	 }
			 else
				 msgList.put(suggestionString, null);
			 
			 if ( result.max_length < r.path.size() ) result.max_length = r.path.size();
			
			 if ( !multiraces ) break;
	 	}
		 
		 // rank advices according to the frequency
		 if ( DataRaceEliminator.choice || ACHSet.choice )
		 {
			 for ( String k : raceSuggestion.keySet() )
			 {
				 MySet<String> suggestions = raceSuggestion.get(k);
				 for ( String s : suggestions )
				 {
					 if ( advices.containsKey(s) )
					 {
						 Integer count = advices.remove(s);
						 advices.put(s, count+1);
					 }
					 else
						 advices.put(s, 1);
				 }
			 }
		 }
		 
		 // show results
		 pw.println("\n\n\nJRF results");
		 MySet<String> raceKinds = raceSet.keySet();
		 
		 long duration = TimeUnit.SECONDS.convert((analysisTime-startTime), TimeUnit.NANOSECONDS);
		 String detectionT = new String((duration/3600)+":"+((duration%3600)/60)+":"+(duration%60));
		 String analysisT=null;
		 int i=0;
		 for ( String r : raceKinds )
		 {
			 pw.println("====================================================== data race #"+(++i));
			 pw.println(r);
			 
			 MyMap<String, String> msgSet = raceSet.get(r);
			 MySet<String> msgList = msgSet.keySet();
			 for ( String msg : msgList )
			 {
				 pw.println(msg);
				 pw.println(msgSet.get(msg));	 
			 }
		 }
		 if ( DataRaceEliminator.choice || ACHSet.choice )
		 {
			 MySet<String> adviceSet = advices.keySet();
			 if ( adviceSet != null && adviceSet.size() > 0 )
			 {
				 pw.println("______________________________________________________ frequency of advice");
			 
				 for ( String a : adviceSet )
				 {
					 pw.println("["+advices.get(a)+"times] "+a);
				 }
			 }
			 
			 duration = TimeUnit.SECONDS.convert((System.nanoTime()-analysisTime), TimeUnit.NANOSECONDS);
			 analysisT = new String((duration/3600)+":"+((duration%3600)/60)+":"+(duration%60));
		 }
		 result.analysis_time = analysisT;
		 result.race_time = detectionT;
		 result.hsize = ListenerHBmanagement.getHBsize();
		 result.hstates = HBSearchListener.getAdditionalStates();
		 
	     pw.println("\n______________________________________________________ statistic");
	     pw.println("JRF explored \""+result.hstates+"\" more states through HB abstraction.");
	     pw.println("Total HB entries = "+result.hsize);
	     pw.println("JRF takes "+detectionT+" to find "+raceKinds.size()+" equivalent races with "+races.size()+" counterexample traces.");
	     pw.println("The maximum counter example path length is \""+result.max_length+"\".");
		 if ( DataRaceEliminator.choice || ACHSet.choice )
		 {
			 pw.println("JRF-E takes "+analysisT+" in "+races.size()+" races analysis.");
		 }
		 return result;
	 }
 
	private static int getTotalNumInstructions(Path path)
	{
		int total = 0;
		for(int i=0; i<path.size(); i++)
			for(Iterator<Step> j = path.get(i).iterator(); j.hasNext();j.next(),total++);
		return total;	   
	}
	  
	/*** Property ****************************************************/
	// This is for timed execution of JRF to avoid indefinite running
	// set the time constraint of JRF, 0 means no constraint
	static final long timeconstraint = JVM.getVM().getConfig().getLong("jrf.timeconstraint");

	// store start time of search
	static long  startTime = System.nanoTime();
	static String errorMessage = new String();
	
	public boolean check (Search search, JVM vm)
	{
		return !isSingleRaceDone() && !isMultiRaceDone() && !isTimeConstraintDone();
	}

	public String getErrorMessage ()
	{
		return errorMessage;
	}
	
	public boolean isSingleRaceDone() // single race mode and found race ?
	{
		if ( !multiraces && races.size()>0 )
		{
			errorMessage += "\n(Data Race)JRF configuration is a single race mode and found a race.\n";
			//errorMessage += ListenerHBmanagement.getHBsize();

			return true;
		}
		return false;
	}
	
	public boolean isMultiRaceDone() // single race mode and found race ?
	{
		if ( multiraces && thresholdraces>0 && races.size()>=thresholdraces )
		{
			errorMessage += "\n(Race Threshold)JRF found "+races.size()+" races that exceeds threshold "+thresholdraces+" races.\n";
			//errorMessage += ListenerHBmanagement.getHBsize();
		
			return true;
		}
		return false;
	}
	
	public boolean isTimeConstraintDone() // time constraint mode and reach time limie ?
	{
		if ( timeconstraint>0 && (System.nanoTime()-startTime)>=timeconstraint )
		{
			long s = TimeUnit.SECONDS.convert((System.nanoTime()-startTime), TimeUnit.NANOSECONDS);
			long limit_s = TimeUnit.SECONDS.convert(timeconstraint, TimeUnit.NANOSECONDS);
			errorMessage += "\n(Time Over)JRF ran for "+(s/3600)+":"+((s%3600)/60)+":"+(s%60)+
							" that exceeds given time constraint "+(limit_s/3600)+":"+((limit_s%3600)/60)+":"+(limit_s%60)+".\n";
			
			return true;
		}
		return false;
	}
	
	// this is called when HBDataRaceException occur
	public static void raceHandler(HBDataRaceException e)
	{	
		JVM jvm = JVM.getVM();
		// single race mode only need one race
		if ( !multiraces && races.size()>0 ) return;
		
		// save the raceSourceInstruction
		Instruction raceInstruction = jvm.getLastInstruction();
			
		Race r = new Race();
		r.raceException = e;
			
		if (  ACHSet.choice )
			r.agents = ListenerHBmanagement.CANDIDATE_AGENTS(e.getMemoryObject());
			
		jvm.updatePath();
		r.path = jvm.getPath().clone();
		r.stepCount = getTotalNumInstructions(r.path);
		r.manifestInstruction = raceInstruction;
		r.pathElements = (PathElementList)pathElements.clone();
		r.snapshot = DataRaceEliminator.getSnapshot();
		r.threadIndex = jvm.getLastThreadInfo().getIndex();
		
		// this is for testing purpose only
		/*
		if ( DataRaceEliminator.choice ) 
			r.counterexampleAnalysis = new DataRaceEliminator().analyzeCounterExample(r.path, 
																						0, 
																						r.threadIndex, 
																						raceSourceInstruction, 
																						r.pathElements, 
																						e.getMemoryObject());
		*/
		races.add(r);	
	}

	/*** SearchListener ****************************************************/

	// this is called only at StateAdvance, so keep in mind that 
	// last instruction is not the cause of this property violation
	public void propertyViolated(Search search) 
	{
		ThreadInfo ti = ThreadInfo.getCurrentThread();
		Log.log("(propertyViolated)"+ti.getName());
	}

	public void searchFinished(Search search) 
	{
		Log.log("(searchFinished)"+"hash entries="+ListenerHBmanagement.getHBsize());
		Log.logResult(ListenerHBmanagement.getHBsize());
	}

	public void searchStarted(Search search)
	{
		// add DataRace as a new property to check
		search.addProperty(this);

		ThreadInfo ti = ThreadInfo.getCurrentThread();
		Log.log("(searchStarted)"+ti.getName());

		// initialize hbset for main thread and its fields
		// note that createInitialObject cannot be used for main thread creation
		ElementInfo thread = JVM.getVM().getDynamicArea().get(ti.getThreadObjectRef());
		ListenerHBmanagement.THREAD_START(thread.toString(), 
				thread.toString(),
				isWithinStaticInit(ti));
		// initialize hbset for main thread' fields 
		for  ( int i=0 ; i < thread.getNumberOfFields() ; ++i )
		{
			FieldInfo fi = thread.getFieldInfo(i);
			Log.log("\t(objectCreated)"+thread.toString()+"."+ 
					fi.getName()+"="+thread.toString()+"."+(i+1)+
					"("+i+"/"+thread.getNumberOfFields()+")");
			
			NEW_OBJECT(ti, thread, fi, i+1, false, isWithinStaticInit(ti)) ;
			// initialize hbset for the memory location already allocated
			if ( fi.isReference() )
			{
				ElementInfo ei ;
				int objRef = thread.getReferenceField(fi);
				if ( fi.isStatic() )
					ei = JVM.getVM().getStaticArea().get(objRef);
				else 
					ei = JVM.getVM().getDynamicArea().get(objRef);

				if ( ei != null ) createInitialObject(JVM.getVM(), ti, ei);
			}
		}

		// initialize hbset for mainThreadGroup and its fields
		ElementInfo threadGroup = JVM.getVM().getDynamicArea().get(thread.getReferenceField("group"));
		createInitialObject(JVM.getVM(), ti, threadGroup);

		// delayed object creation
		Iterator<ElementInfo> eit = delayed_object_creation.iterator();
		while( eit.hasNext() )
		{
			ElementInfo ei = eit.next();
			createObject(JVM.getVM(), ThreadInfo.getCurrentThread(), ei);
		}

		// delayed class creation
		Iterator<ClassInfo> cit = delayed_class_creation.iterator();
		while( cit.hasNext() )
		{
			ClassInfo ci = cit.next();
			createClass(JVM.getVM(), ThreadInfo.getCurrentThread(), ci);
		}

		Initialized = true;
		Log.log("(searchStarted) finishes initializing..."+ti.getName());
	}

	private void createInitialObject(JVM jvm, ThreadInfo ti, ElementInfo ei)
	{
		// set allocation location for initial objects
		Key o = new Key(ei.toString(), 0);
		o.setLocation(HBDataRaceException.getSourceLocation());
		
		// create HBentry for object
		Log.log("\t(objectCreatedInitially)"+ei.toString());
		NEW_OBJECT(ti, ei, null, 0, false, isWithinStaticInit(ti));
		// create HBentries for fields
		for  ( int i=0 ; i < ei.getNumberOfFields() ; ++i )
		{
			FieldInfo fi = ei.getFieldInfo(i);
			Log.log("\t(objectCreatedInitially)"+ei.toString()+"."+ fi.getName()+"="+ei.toString()+"."+
					(i+1)+"("+i+"/"+ei.getNumberOfFields()+")");
			NEW_OBJECT(ti, ei, fi,i+1, false, isWithinStaticInit(ti));

			// non-null reference field needs additional object creation for the memory assigned to it 
			if ( fi.isReference() )
			{
				ElementInfo fei;
				int objRef = ei.getReferenceField(fi);
				if ( fi.isStatic() )
					fei = JVM.getVM().getStaticArea().get(objRef);
				else 
					fei = JVM.getVM().getDynamicArea().get(objRef);
				//Log.log("thread="+ti+",fi="+fi+",objRef="+ei.getReferenceField(fi)+",fei="+fei);

				if ( fei != null ) createInitialObject(jvm, ti, fei);
			}
		}
		
		// if object is an array, we need HBentries for each index
		// "class name@object number[index]" will be used as a key 
		if ( ei.isArray() )
		{
			Log.log("\t(objectCreatedInitially)"+ei.toString()+"["+"LAZY"+"]");
			NEW_OBJECT(ti, ei, null, AbstractArrayImpl.LAZY, false, isWithinStaticInit(ti));
		}
		
	}

	// store hb related information here
	public void stateStored(Search search) 
	{
		
		int stateNum = search.getStateNumber();
		Log.log("(stateStored)"+stateNum+","+search.getDepth());
		//ListenerHBmanagement.STATE_STORE();

		// store current state into "store"
		State newStore = new State(	(PathElementList)(pathElements.clone()),
									(MyMap<Integer,Integer>)(statePathIndex.clone()),
									ListenerHBmanagement.STATE_STORE());
		//if ( search.getTransition() != null )
		//System.out.println("### STORE("+search.getTransition().getThreadIndex()+" " +search.getStateNumber()+" "+search.getDepth()+")="+newStore.hbData);	

		ArrayList<State> list = store.remove(stateNum);
		if ( list == null ) list = new ArrayList<State>();
		
		if ( ListenerHBmanagement.is_FIFO_Order() )
			// in case of BFS
			list.add(0, newStore);
		else
			// in case of DFS
			list.add(newStore);
		
		//System.out.println("***** stored at "+stateNum+"="+list.size()+":"+list.get(0).hbData+", "+(list.size()>1?list.get(1).hbData:"end"));
		store.put(stateNum, list);

	}

	// restore only need to remove states from store
	public void stateRestored(Search search) 
	{
		int stateNum = search.getStateNumber();
		Log.log("(stateRestored)"+stateNum+","+search.getDepth());
		

		// restore values stored by listener
		ArrayList<State> oldlist = store.remove(stateNum); 
		State old = oldlist.remove(oldlist.size()-1);//, search.getDepth()));	
		if ( oldlist.size() > 0 ) store.put(stateNum, oldlist);
		//System.out.println("***** after restored at "+stateNum+"="+oldlist.size()+":"+(oldlist.size()>0?oldlist.get(0).hbData:"")+", "+(oldlist.size()>1?oldlist.get(1).hbData:"end"));
			
		pathElements = old.pathElements;
		statePathIndex = old.statePathIndex;
		ListenerHBmanagement.STATE_RESTORE(old.hbData);
//System.out.println("### RESTORE("+search.getTransition().getThreadIndex()+" " +search.getStateNumber()+" "+search.getDepth()+")="+old.hbData);	
	}


	// store information related with path and within static initializer nested depth
	public void stateAdvanced(Search search) {

		Log.log("(stateAdvanced) STATE "+search.getStateNumber()+", DEPTH "+search.getDepth()+", THREAD "+search.getTransition().getThreadIndex());

		ListenerHBmanagement.STATE_ADVANCE(search.getStateNumber()); //, (HBSearch)search);

		// store current path memory objects
		if ( !statePathIndex.containsKey(new Integer(search.getStateNumber())) )
			statePathIndex.put(new Integer(search.getStateNumber()), new Integer(pathElements.size()));	
	}

	// state backtrack needs to recover previous value of path and within static initializer
	// if state number is not match, we might need to backtrack more than one state
	public void stateBacktracked(Search search) {

		// backtrack to specified state number
		if ( search.getStateNumber() == Log.EMPTY )
		{
			//Log.log("!!!SKIP!!! (stateBacktracked) STATE "+search.getStateNumber()+", DEPTH "+search.getDepth()+
			//		(search.getTransition()==null?"":(", THREAD "+search.getTransition().getThreadIndex())));
		}
		else
		{
			Log.log("(stateBacktracked) STATE "+search.getStateNumber()+", DEPTH "+search.getDepth()+", THREAD "+search.getTransition().getThreadIndex());
			ListenerHBmanagement.STATE_BACKTRACK(search.getStateNumber());

			// restore state information also
			// need to restore other informations???
			int index = statePathIndex.get(new Integer(search.getStateNumber())).intValue();
			//Log.log("((STATE_BACKTRACKED)) get "+search.getStateNumber()+"->"+index+", current path="+pathElements.size());
			for (int i= pathElements.size(); i > index ; --i )
				pathElements.remove(i-1);
		}
	}

	public void stateProcessed (Search search)
	{
		Log.log("(stateProcessed)"+search.getStateNumber()+","+search.getDepth());
	}


	/*** VMListener *******************************************************/
	// this is dealing with threadStart HB
	public void threadStarted(JVM jvm) {

		if ( !Initialized ) return;

		ThreadInfo c = jvm.getLastThreadInfo();
		ThreadInfo p = jvm.getCurrentThread();
		Log.log("\t\t(threadStarted)"+c.getName());

		ListenerHBmanagement.THREAD_START(getThreadObject(p), 
				getThreadObject(c),
				isWithinStaticInit(p));
	}

	//long value_before;
	
	// need to store the current instruction before other listener routine is invoked
	public void executeInstruction(JVM jvm)
	{
		Instruction insn = jvm.getLastInstruction();		
		if ( !Initialized ) return;
		HBDataRaceException.setSourceLocation(insn);
	}
	
	//
	// note that stack information is already changed by executing the instruction, so doesn't have
	// arguments on its top, but has return value of the instruction if presents
	//
	public void instructionExecuted(JVM jvm) {
		if ( !Initialized ) return;

		Instruction insn = jvm.getLastInstruction();
		ThreadInfo ti = jvm.getLastThreadInfo();
		
	
			Log.logInstruction("\t\t\tINSTRUCTION( completed="+insn.isCompleted(ti)+"("+insn.getSourceLine()+"):"+insn+"##"+JRFListener.isWithinStaticInit(ti)+"##");

			if (insn instanceof FieldInstruction) {
//	if ( insn.isCompleted(ti) )
//	{
				FieldInstruction finsn = (FieldInstruction)insn;
				FieldInfo fi = finsn.getFieldInfo();
				//ThreadInfo ti = jvm.getLastThreadInfo();

				ElementInfo ei= null;
				int fn = -1;
				if (insn instanceof StaticFieldInstruction) { 
					ei = jvm.getClassReference(fi.getClassInfo().getName());
					fn = fi.getFieldIndex()+1;
					assert(fn!=0):"Cannot decide Field Number of \""+fi.getName()+"\" from class "+ei;

					// save to path
					assert(finsn.getLastElementInfo()==ei);
					if ( DataRaceEliminator.choice ) {  pathElements.add(new Key(ei.toString(), fn)); } 
					
					if ( insn instanceof GETSTATIC )
					{
						// getstatic Class.field =  (read of Class.field)
						READ(ti, ei, fi, fn, fi.isVolatile(), isWithinStaticInit(ti));
					}
					else // insn instanceof PUTSTATIC 
					{
						// putstatic Class.field =  (write of Class.field)
						/* redundant update 
						if ( fi.isVolatile() || ((PUTSTATIC)insn).getLastValue() != value_before )
						*/
						WRITE(ti, ei, fi, fn, fi.isVolatile(), isWithinStaticInit(ti));
					}

				} else { // instance field

					ei = ((InstanceFieldInstruction)insn).getLastElementInfo();
					fn = fi.getFieldIndex()+1;
					assert(fn!=0) :"Cannot decide Field Number of \""+fi.getName()+"\" from "+ei;

					// save to path
					assert(finsn.getLastElementInfo()==ei);
					if ( DataRaceEliminator.choice ) { pathElements.add(new Key(ei.toString(), fn)); } 
					
					if ( insn instanceof GETFIELD )
					{
						// getfield Object.field = (read of Object) (read of Object.field)
						READ(ti, ei, null, 0, false, isWithinStaticInit(ti));

						READ(ti, ei, fi, fn, fi.isVolatile(), isWithinStaticInit(ti));
					}
					else // insn instanceof PUTFIELD 
					{

						// putfield Object.field = (read of Object) (read of Object.field)
						READ(ti, ei, null, 0, false, isWithinStaticInit(ti));
						/* redundant update 
						if ( fi.isVolatile() || ((PUTFIELD)insn).getLastValue() != value_before )
						 */
							WRITE(ti, ei, fi, fn, fi.isVolatile(), isWithinStaticInit(ti));
					}
				}
//	}
//	else if ( DataRaceEliminator.choice ) {  pathElements.add(null); } 
			}
			// handle array instructions
			else if ( insn instanceof ArrayLoadInstruction )
			{
//	if ( insn.isCompleted(ti) )
//	{			
				//ThreadInfo ti = jvm.getLastThreadInfo();
				ArrayLoadInstruction ai = (ArrayLoadInstruction)insn;
				ElementInfo ei = ti.getElementInfo(ai.getArrayRef(ti));//getStoredElementInfo();
				int index = ai.getIndex(ti);//getStoredIndex();
				
				// save to path
				if ( DataRaceEliminator.choice ) { pathElements.add(new Key(ei.toString(), index+1)); }

				// array element is not volatile always
				READ(ti, ei, null, 0, false, isWithinStaticInit(ti));
				READ(ti, ei, null, index+1, false, isWithinStaticInit(ti));
//	}
//	else if ( DataRaceEliminator.choice ) {  pathElements.add(null); } 
			}
			else if ( insn instanceof ArrayStoreInstruction )
			{
//	if ( insn.isCompleted(ti) )
//	{
				ArrayStoreInstruction ai = (ArrayStoreInstruction)insn;
				ElementInfo ei = ti.getElementInfo(ai.getArrayRef(ti));//getStoredElementInfo();
				int index = ai.getIndex(ti);//getStoredIndex();
				
				// save to path
				if ( DataRaceEliminator.choice ) { pathElements.add(new Key(ei.toString(), index+1)); }
				
				
				// array element is not volatile always
				READ(ti, ei, null, 0, false, isWithinStaticInit(ti));
				WRITE(ti, ei, null, index+1, false, isWithinStaticInit(ti));
//	}		
//	else if ( DataRaceEliminator.choice ) {  pathElements.add(null); } 
			}
			else if ( (insn instanceof NEW ) 
					|| (insn instanceof NEWARRAY ) 
					|| (insn instanceof ANEWARRAY ) 
					|| (insn instanceof MULTIANEWARRAY ) )
			{
				if ( insn.isCompleted(jvm.getLastThreadInfo()) )
				{
					int objref = ti.peek();
					ElementInfo ei = jvm.getDynamicArea().get(objref);
					Key o = new Key(ei.toString(), 0);
					
					// save to path
					if ( DataRaceEliminator.choice ) { pathElements.add(o); } 
					
					// save source position of allocation
					o.setLocation(HBDataRaceException.getSourceLocation());
				} 
				else
					if ( DataRaceEliminator.choice ) { pathElements.add(null); } 
			}
			else if ( insn instanceof InvokeInstruction )
			{
				InvokeInstruction iinsn = (InvokeInstruction)insn;
				MethodInfo m = iinsn.getInvokedMethod();
		
				if ( m== null )
					m = ClassInfo.getResolvedClassInfo(iinsn.getInvokedMethodClassName()).getMethod(iinsn.getInvokedMethodName(), true);
				
				if ( m.isSynchronized() )
				{
					ElementInfo ei = jvm.getLastElementInfo();
					if ( ei== null )
						ei = m.getClassInfo().getStaticElementInfo();
					
					// save to path
					if ( DataRaceEliminator.choice ) { pathElements.add(new Key(ei.toString(), 0)); }
				}
				else
				{
					// save to path
					if ( DataRaceEliminator.choice ) { pathElements.add(null); }
				}
			}
			else if ( insn instanceof ReturnInstruction )
			{
				MethodInfo m = ((ReturnInstruction)insn).getMethodInfo();
	
				if ( m.isSynchronized() ) 
					// invoke to synchronized method, we need to save memory object for this
				{
					ElementInfo ei = jvm.getLastElementInfo();
					// save to path
					if ( DataRaceEliminator.choice ) { pathElements.add(new Key(ei.toString(), 0)); } 
				}
				else
				{
					// save to path
					if ( DataRaceEliminator.choice ) { pathElements.add(null); } 
				}
			}
			else if ( insn instanceof LockInstruction )
			{
				LockInstruction mi = (LockInstruction)insn;
				// save to path
				if ( DataRaceEliminator.choice ) { pathElements.add(new Key(mi.getLastLockElementInfo().toString(), 0)); } 
			}
			else // all other case
			{
				// save to path
				if ( DataRaceEliminator.choice ) { pathElements.add(null); } 
			}
	}

	// this is called by gov.nasa.jpf.jvm.ThreadInfo
	public void objectLocked(JVM jvm)
	{
		if ( !Initialized ) return;
		
		ElementInfo ei = jvm.getLastElementInfo();
		ThreadInfo ti = jvm.getLastThreadInfo();
		
			// lock object is 
			// 		1) read of object  (?? is this needed???)
			//		2) acquire of object
			Log.log("\t(objectLocked)"+ei.toString()+" by "+getThreadObject(ti)+".");
			READ(ti, ei, null, 0, false, isWithinStaticInit(ti));
			ListenerHBmanagement.LOCK(getThreadObject(ti), ei.toString(), isWithinStaticInit(ti)); 
	}

	// this is called by gov.nasa.jpf.jvm.ThreadInfo 
	public void objectUnlocked(JVM jvm)
	{
		if ( !Initialized ) return;

			ElementInfo ei = jvm.getLastElementInfo();
			ThreadInfo ti = jvm.getLastThreadInfo();	
			Log.log("\t(objectUnlocked)"+ei.toString()+" by "+getThreadObject(ti)+".");
			// unlock object is 
			// 		1) read of object  (?? is this needed???)
			//		2) release of object
			READ(ti, ei, null, 0, false, isWithinStaticInit(ti));
			ListenerHBmanagement.UNLOCK(getThreadObject(ti), ei.toString(), isWithinStaticInit(ti)); 
	}



	/**
	 * object was garbage collected (after potential finalization)
	 */
	public void objectReleased (JVM jvm)
	{
		ElementInfo ei = jvm.getLastElementInfo();
		if ( !Initialized ) 
		{
			// if object is released before search start, 
			// there is no need to create HBentry
			assert(ei!=null):"ElementInfo is null";
			delayed_object_creation.remove(ei);
			return;
		}
		Log.log("\t(objectReleased)"+ei.toString());
	}

	// this is called by gov.nasa.jpf.jvm.DynamicArea newObject or newArray
	
	public void objectCreated(JVM jvm)
	{
		ElementInfo ei = jvm.getLastElementInfo();
		// delay object creation if search is not started
		if ( !Initialized ) 
		{
			assert(ei!=null):"ElementInfo is null";
			delayed_object_creation.add(ei);
			return;
		}
		// go ahead and create HBentry
		createObject(jvm, jvm.getLastThreadInfo(), ei);
	}
    
	
	// create instance object by given thread ti
	private void createObject(JVM jvm, ThreadInfo ti, ElementInfo ei)
	{
		// create HBentry for object
		Log.log("\t(objectCreated)"+ei.toString());
		
		// set allocation location for initial objects
		Key o = new Key(ei.toString(), 0);
		if ( o.getLocation() == null )
			o.setLocation(HBDataRaceException.getSourceLocation());
		
		NEW_OBJECT(ti, ei, null, 0, false, isWithinStaticInit(ti));
		
		// create HBentries for fields
		for  ( int i=0 ; i < ei.getNumberOfFields() ; ++i )
		{
			FieldInfo fi = ei.getFieldInfo(i);

			Log.log("\t(objectCreated)"+ei.toString()+"."+ fi.getName()+"="+ei.toString()+"."+(i+1)+"("+i+"/"+ei.getNumberOfFields()+")");
			NEW_OBJECT(ti, ei, fi, i+1, fi.isVolatile(), isWithinStaticInit(ti));
		}
		/* ArrayAbstraction for better performance */
		/* We will create hb entry for each array element only when its invalidation results in different agents */
		
		// if object is an array, we need HBentries for each index
		// "class name@object number[index]" will be used as a key 
		if ( ei.isArray() )
		{
			Log.log("\t(objectCreated)"+ei.toString()+"."+"LAZY");
			NEW_OBJECT(ti, ei, null, AbstractArrayImpl.LAZY, false, isWithinStaticInit(ti));
		}
		
	}
    

	// called by gov.nasa.jpf.jvm.ClassInfo initialize
	public void classLoaded(JVM jvm) 
	{
		ClassInfo ci = jvm.getLastClassInfo();
		ThreadInfo ti = jvm.getCurrentThread();
		
		Log.log("\t(classLoaded)"+ci.getName()+" by "+(ti==null?ti:jrf.hbset.util.Utility.getThreadObject(ti)));
		
		// if search is not started yet, delay HBentry creation after that
		if ( !Initialized ) 
		{
			delayed_class_creation.add(ci);
			return;
		}
		// after search start, so go ahead and create HBentry
		createClass(jvm, ti, ci);
	}

	// create class object by given thread ti
	// class object creation is done at class load time, so it will neither be 
	// canceled by stateBacktrack nor be replayed by next stateAdvance
	private void createClass(JVM jvm, ThreadInfo ti, ClassInfo ci)
	{	
		// create HBentry for class object
		Log.log("\t(classCreated)"+ci.getName());

		// set allocation location for initial objects
		Key o = new Key(ci.getName(), 0);
		if ( o.getLocation() == null )
			o.setLocation(HBDataRaceException.getSourceLocation());
		
		NEW_CLASS(ti, ci, null, 0, false, true);

		// create HBentry for class object static fields
		for  ( int i=0 ; i < ci.getNumberOfStaticFields() ; ++i )
		{
			FieldInfo fi = ci.getStaticField(i);
			Log.log("\t(classFieldCreated)"+ci.getName()+"."+ fi.getName()+"="+ci.getName()+"."+(i+1)+"("+i+"/"+ci.getNumberOfStaticFields()+")");
			NEW_CLASS(ti, ci, fi, i+1, fi.isVolatile(), true);
		}
	}
}
