package jrf.eliminator;


import gov.nasa.jpf.jvm.ClassInfo;
import gov.nasa.jpf.jvm.JVM;
import gov.nasa.jpf.jvm.Path;
import gov.nasa.jpf.jvm.FieldInfo;
import gov.nasa.jpf.jvm.MethodInfo;
import gov.nasa.jpf.jvm.Transition;
import gov.nasa.jpf.jvm.Step;
import gov.nasa.jpf.jvm.bytecode.*;
import gov.nasa.jpf.util.Left;

import java.util.Iterator;
import java.util.ArrayList;
import jrf.hbset.util.*;

public class DataRaceEliminator {
	public static final boolean choice = JVM.getVM().getConfig().getBoolean("jrf.eliminator");
	
	// Stores the set of happens-before edges as a pair of instruction indices in Range objects
	// The instruction indices are determined based on the execution order on the counter-example path 
	private ArrayList<Range> happensBeforeEdges = null;    

	// Stores the index of the instruction in which the data race has been manifested
	private int manifestingInstructionNo = -1;
	
	// Stores the index of the instruction that is the source of the data race 
	private int sourceInstructionNo = -1;

	private  boolean isWrite(Instruction insn)
	{
		return (insn instanceof PUTFIELD || insn instanceof PUTSTATIC || insn instanceof ArrayStoreInstruction || insn instanceof NEW); 

	}


	private  boolean isRead(Instruction insn)
	{
		return (insn instanceof GETFIELD || insn instanceof GETSTATIC || insn instanceof ArrayLoadInstruction ); 
	}



	private  Key getLockedMemoryLocation(Instruction insn, PathElementList memoryObjects,int pathIndex)
	{
		if (insn == null) return null;
		return memoryObjects.get(pathIndex);
	}



	/* Should also handle aliasing */
	private  Key getMemoryLocation(Instruction insn, PathElementList memoryObjects, int pathIndex)
	{
		if (insn == null) return null;
		return memoryObjects.get(pathIndex);
	}


	 class KeyLine {
		Key key;
		String sourceLine;
		public KeyLine(Key key,String sourceLine)
		{
			this.key = key;
			this.sourceLine = sourceLine; 
		}
	}	 

	 class InstructionInfo {
		Instruction instr;
		int index;
		int threadId;
		public InstructionInfo(Instruction instr, int index,int threadId)
		{
			this.instr = instr;
			this.index = index;
			this.threadId = threadId;
		}

	}


	private  int getTotalNumInstructions(Path path)
	{
		int total = 0;
		for(int i=0; i<path.size(); i++)
			for(Iterator<Step> j = path.get(i).iterator(); j.hasNext();j.next(),total++);
		return total;	   
	}



	private  ArrayList<KeyLine> findMemoryLocationsToSynchronizeOn(Path path, InstructionInfo sourceInstr,int stepCount, PathElementList memoryObjects)
	{
	        int sourceIndex = sourceInstr.index;
	        int sourceTID = sourceInstr.threadId;
		ArrayList<KeyLine> result = new ArrayList<KeyLine>();        
		int count = 0;
		int index = getTotalNumInstructions(path);
		boolean sourceReached = false;
		for(int i=path.size()-1; !sourceReached;i--)
		{
			Transition t =path.get(i);
			ArrayList<Step> vv=reverse(t);
			for(int vindex=0; vindex<vv.size(); vindex++, index--,count++)
			{
				if (index == sourceIndex) 
				{
					sourceReached = true;
					break; 
				}
				if (count <= stepCount) continue;
				Step s = vv.get(vindex);
				Instruction ins = s.getInstruction();
				if (ins instanceof MONITOREXIT && t.getThreadIndex() == sourceTID)
					result.add(new KeyLine(getMemoryLocation(ins,memoryObjects,index-1),
								"\""+ins.getSourceLine().trim() +"\" at "+ ins.getFileLocation().trim()));           
				else if (ins instanceof RETURN && t.getThreadIndex() == sourceTID)
				{
					MethodInfo m = ins.getMethodInfo();
					if (m.isSynchronized())
					{
						result.add(new KeyLine(getLockedMemoryLocation(ins,memoryObjects,index-1),
								"\""+ins.getSourceLine().trim() + "\" at " + ins.getFileLocation().trim()));
					}   
				}   
			}                                    
		}
		return result;
	}

	private  ArrayList<Key> findMemoryLocationsToChangeVolatile(Path path, InstructionInfo sourceInstr,int stepCount, PathElementList memoryObjects, Range target)
	{
	    int sourceIndex = sourceInstr.index;
	    int sourceTID = sourceInstr.threadId;
		ArrayList<Key> result = new ArrayList<Key>();        
	        
	    // find non-volatil read-write pair in the Range(source, manifest)   
	    // and compute possible HBedges by changing it to volatile
	    MyMap<Key, ArrayList<Range>> newHBedges = new MyMap<Key, ArrayList<Range>>();
	    MyMap<Key, ArrayList<InstructionInfo>> readLists = new MyMap<Key, ArrayList<InstructionInfo>>();
	    
		int count = 0;
		int index = getTotalNumInstructions(path);
		boolean sourceReached = false;
		for(int i=path.size()-1; !sourceReached;i--)
		{
			Transition t =path.get(i);
			ArrayList<Step> vv=reverse(t);
			for(int vindex=0; vindex<vv.size(); vindex++, index--,count++)
			{
				if (index == sourceIndex) 
				{
					sourceReached = true;
					break; 
				}
				if (count <= stepCount) continue;
				Step s = vv.get(vindex);
				Instruction ins = s.getInstruction();
				
				if ( isRead(ins) && !isVolatile(ins) ) //&& t.getThreadIndex() == raceTID )
				{
					Key k = getMemoryLocation(ins,memoryObjects,index-1);
					if ( k == null ) continue; // taking care of isCompleted change in JRFlistener
					ArrayList<InstructionInfo> list;
					if ( readLists.containsKey(k) ) list = readLists.remove(k);
					else list = new ArrayList<InstructionInfo>();
					list.add(new InstructionInfo(ins, index, t.getThreadIndex()));
					readLists.put(k, list);
				}
				else if ( isWrite(ins) && !isVolatile(ins) )
				{
					Key k = getMemoryLocation(ins,memoryObjects,index-1);
					if ( k == null ) continue; // taking care of isCompleted change in JRFlistener
					ArrayList<InstructionInfo> list = readLists.remove(k);
					if ( list == null ) continue;
					
					ArrayList<Range> edges;
					if ( newHBedges.containsKey(k) ) edges = newHBedges.remove(k);
					else edges = new ArrayList<Range>();
					
					for ( InstructionInfo readI : list )
					{
						edges.add(new Range(index, ins, t.getThreadIndex(), readI.index, readI.instr, readI.threadId));
					}
					newHBedges.put(k, edges);
				}
			}
			
			// check if the change could give the hb order to target
		}
		
		MySet<Key> keySet = newHBedges.keySet();
		for ( Key k : keySet )
		{
			ArrayList<Range> additional = newHBedges.get(k);
			additional.addAll(target.inclusive(happensBeforeEdges));
			if ( isHBOrdered(target, additional) )
				result.add(k);
		}
		return result;
	}
	
	private boolean isHBOrdered(Range target, ArrayList<Range> hbEdges)
	{
		if ( target.highTid == target.lowTid ) return true; // program order
		ArrayList<Range> internal = target.inclusive(hbEdges);
		
		for ( Range r : internal )
		{
			if ( r.low == target.low && r.high == target.high ) return true; // has target hb
			Range lower = new Range(target.low, target.lowIns, target.lowTid, r.low, r.lowIns, r.lowTid);
			Range higher = new Range(r.high, r.highIns, r.highTid, target.high, target.highIns, target.highTid);
			if ( isHBOrdered(lower, internal) && isHBOrdered(higher, internal) ) return true;
		}
		return false;
	}
	
	private  ArrayList<Step> reverse(Transition t)
	{ 
		ArrayList<Step> result = new ArrayList<Step>();
		for(Iterator<Step> ss=t.iterator(); ss.hasNext();)
			result.add(0,ss.next());
		return result;   
	}

	// After the data race has been manifested stepCount many steps have been executed 
	// before termination  
	private  InstructionInfo findRaceSource(Key memLoc,Path path, int stepCount, PathElementList memoryLocations)
	{
		InstructionInfo result = new InstructionInfo(null,-1,-1);
		boolean found = false;

		int count =0; 
		int i=path.size()-1;
		int index=getTotalNumInstructions(path);

		while (!found && i>=0)
		{
			Transition t =path.get(i);
			//for(Iterator<Step> ss=t.iterator(); ss.hasNext();index--,count++)
			ArrayList<Step> vv=reverse(t);
			for(int vindex=0; vindex<vv.size(); vindex++, index--,count++)
			{
				Step s = vv.get(vindex);
				// Should skip analyzing the data race manifesting statement and 
				// all the steps executed after it
				if (count <= stepCount) continue; 
				Instruction ins = s.getInstruction();
				//System.out.println("findRaceSource "+ins + ": " + ins.getFileLocation() + " \"" + ins.getSourceLine());
				Key m = getMemoryLocation(ins,memoryLocations,index-1);

				if (m == null) continue;	
				if (m.equals(memLoc) && isWrite(ins)) 
				{
					found = true;
					result = new InstructionInfo(ins,index,t.getThreadIndex());
					break;
				}	           
			}
			i--;
		}		
		return result;
	}

	private static enum instructionType {READ_VOLATILE, ACQUIRE_LOCK, WRITE_VOLATILE, RELEASE_LOCK, OTHER };


	private  boolean matches(instructionType i1, instructionType i2)
	{
		return ((i1 == instructionType.READ_VOLATILE && i2 == instructionType.WRITE_VOLATILE) || 
				(i2 == instructionType.READ_VOLATILE && i1 == instructionType.WRITE_VOLATILE) ||
				(i1 == instructionType.ACQUIRE_LOCK && i2 == instructionType.RELEASE_LOCK) ||
				(i2 == instructionType.ACQUIRE_LOCK && i1 == instructionType.RELEASE_LOCK)) ; 
	}

	private  instructionType getInstructionType(Instruction insn)
	{
		if (insn instanceof FieldInstruction)
		{
			FieldInstruction finsn = (FieldInstruction)insn;
			FieldInfo fi = finsn.getFieldInfo();
			if (fi.isVolatile() && isRead(insn))
				return instructionType.READ_VOLATILE;
			else if (fi.isVolatile() && isWrite(insn))
				return instructionType.WRITE_VOLATILE;
			else return instructionType.OTHER;                 
		}
		else if (insn instanceof MONITORENTER)
			return instructionType.ACQUIRE_LOCK;
		else if (insn instanceof MONITOREXIT)
			return instructionType.RELEASE_LOCK;
		else return instructionType.OTHER;   
	}


	 class Range {
		int low;
		int high;
		Instruction lowIns;
		Instruction highIns;
		int lowTid;
		int highTid;

		private Range(int l, Instruction lIns, int ltid, int h, Instruction hIns,int htid)
		{
			low = l;
			lowIns = lIns;
			high = h;
			highIns = hIns;
			lowTid = ltid;
			highTid = htid;
		}

		private boolean intersects(Range r)
		{
			return ((low <= r.low && r.low <= high) || (low <= r.high && r.high <= high)
					|| (r.low <= low && low <= r.high) || (r.low <= high && high <= r.high));
		}

        private boolean includes(Range r)
        {
        	return (low <= r.low && high >= r.high);   
        }

        public boolean before(Range r)
        {
        	return (low < r.low);
        }
                 
        public String toString()
		{
			return "InstructionId: [" + low +"," + high +"] ThreadId: [" + lowTid + "," + highTid +"]";
		}
        
        public ArrayList<Range> inclusive(ArrayList<Range> lists)
        {
        	ArrayList<Range> result = new ArrayList<Range>();
        	for ( Range r : lists)
        		if ( includes(r) ) result.add(r);
        	return result;
        }
	}

        private  InstructionInfo getInstruction(int index, Path path)
        {
           int count = getTotalNumInstructions(path);
           for(int i=path.size()-1; i>=0; i--)
              for(int j = path.get(i).getStepCount() - 1; j>=0;j--,count--)
              {
                  Step s= path.get(i).getStep(j);
                  if (count == index)
                     return new InstructionInfo(s.getInstruction(),index,path.get(i).getThreadIndex());
              }
           return null;   
        }

	private  ArrayList<Range> findHBEdges(Path path, PathElementList memoryObjects)
	{
		if (happensBeforeEdges != null) return happensBeforeEdges;

		ArrayList<Range> hbEdges = new ArrayList<Range>(); 
		instructionType acquireType;
		instructionType releaseType;
		int acquireI;
		int releaseI;
		int acquireTid;
		int releaseTid;
		Instruction acquireIns;
		Instruction releaseIns;
		int count = getTotalNumInstructions(path);
		for(int i=path.size()-1;i>=0;i--)
		{
			for(int j = path.get(i).getStepCount() - 1 ; j>=0; j--, count--)
			{
				Step s1 = path.get(i).getStep(j);
				Instruction ins1 = s1.getInstruction();
				if (isAcquire(ins1))
				{
					acquireIns = ins1;
					acquireI = count;
					acquireTid = path.get(i).getThreadIndex();
					acquireType = getInstructionType(ins1);          

					int k = j;
					int m = i;
					int countCopy = count;
					boolean foundRelease = false;
					boolean foundObjectCreationSite = false; 
					for(; m>=0 && !foundRelease && !foundObjectCreationSite; )
					{
						for(; k>=0; k--, countCopy--)
						{
							Step s2 = path.get(m).getStep(k); 
							Instruction ins2 = s2.getInstruction();
							if (isRelease(ins2))
							{
								releaseIns = ins2;
								releaseI = countCopy;
								releaseType = getInstructionType(ins2);
								releaseTid = path.get(m).getThreadIndex();
								Key mem_acq = getMemoryLocation(acquireIns,memoryObjects,acquireI-1);   
								Key mem_rel = getMemoryLocation(releaseIns,memoryObjects,releaseI-1);
								//System.out.println(acquireType + " " + releaseType); 
								if (matches(acquireType,releaseType) && (mem_acq == null || mem_rel == null || mem_acq.equals(getMemoryLocation(releaseIns,memoryObjects,releaseI-1))))
								{
									hbEdges.add(new Range(releaseI,releaseIns,releaseTid,acquireI,acquireIns,acquireTid));
									//System.out.println("rI=" + releaseI + "rIns=" + releaseIns + "rTid="+ releaseTid + "aI=" + acquireI + "aIns="+acquireIns + "aTid="+acquireTid);
									foundRelease = true;
									if (ins2 instanceof NEW)
										foundObjectCreationSite = true;  
									break; 
								}   
							} 	                  
						}
						m--;
						if (m>=0)
							k = path.get(m).getStepCount() - 1;
					} 
				} 
			}    

		}
		happensBeforeEdges = hbEdges;
		return hbEdges;

	}

	private  boolean isVolatile(Instruction insn)
	{
		if (insn instanceof FieldInstruction)
		{
			FieldInstruction finsn = (FieldInstruction)insn;
			FieldInfo fi = finsn.getFieldInfo();
			return fi.isVolatile();           
		}
		return false;
	}


	private  boolean isAcquire(Instruction insn)
	{
		if (insn instanceof FieldInstruction)
		{
			FieldInstruction finsn = (FieldInstruction)insn;
			FieldInfo fi = finsn.getFieldInfo();
			return fi.isVolatile() && isRead(insn);           
		}
		else if (insn instanceof MONITORENTER)
			return true;
		return false;
	}

	private  boolean isRelease(Instruction insn)
	{
		if (insn instanceof FieldInstruction)
		{
			FieldInstruction finsn = (FieldInstruction)insn;
			FieldInfo fi = finsn.getFieldInfo();
			return fi.isVolatile() && isWrite(insn);           
		}
		else if (insn instanceof MONITOREXIT)
			return true;
		return false;
	}

	private  boolean isArray(Instruction insn)
	{
		return (insn instanceof ArrayInstruction);
	}
	
	public  ArrayList<String> analyzeCounterExample( 
			Path path,
			int stepCounter,
			int threadIndex,
			Instruction errorInstr, 
			PathElementList memoryObjects, Key memoryLocationKey)
	{
		manifestingInstructionNo = stepCounter;
		
		ArrayList<String> message = new ArrayList<String>();
		Instruction lastInstruction = errorInstr;	

		// compute total number of steps in current path
		int step=0;
		Iterator<Transition> it = path.iterator();
		while ( it.hasNext() )
		{
			Transition t = it.next();
			step += t.getStepCount();
		}
		
		// find source statement
		InstructionInfo raceSourceI = findRaceSource(memoryLocationKey,path,stepCounter,memoryObjects); 
		sourceInstructionNo = raceSourceI.index;
		
		if (raceSourceI.instr == null) message.add("data race source statement   : null");
		else {message.add("data race source statement : \""+raceSourceI.instr+
				"\" at "+raceSourceI.instr.getFileLocation().trim()+
				"  : \""+raceSourceI.instr.getSourceLine().trim()+"\"");}
		
		message.add("data race manifest statement : \""+errorInstr+
				"\" at "+errorInstr.getFileLocation().trim()+
				": \""+errorInstr.getSourceLine().trim()+"\"\n");
		
		if (raceSourceI.instr == null) return message;
		
		if ( raceSourceI.instr.getMethodInfo()!= null && raceSourceI.instr.getMethodInfo().isCtor() && errorInstr instanceof GETFIELD )
		{
			Key obj = new Key(memoryLocationKey, 0);
			message.add("Unsafe publication of object "+obj.toNameStringWithLocation()+"\n\t"
					+" caused the race on the field \""+memoryLocationKey.toNameString()+"\".\n");
		}
		
		
		int index2 = getTotalNumInstructions(path)-1; 
		
		//
		// SUGGESTION (1) change to volatile or using atomic array
		//
		if (isWrite(raceSourceI.instr) && (isRead(lastInstruction) || isWrite(lastInstruction)))
		{
			
			if ( isArray(lastInstruction) )
			{
				message.add("Change the type of the array field \""+memoryLocationKey.toNameStringWithLocation()+"\" to use java.util.concurrent.atomic package.");
				/*message.add("Change the type of the array field in line "+
						lastInstruction.getFileLocation().trim()  +
						": \""+lastInstruction.getSourceLine().trim()+"\" to use java.util.concurrent.atomic package.");
						*/
			}
			else if (!isVolatile(lastInstruction) && memoryLocationKey.isField() )
				message.add("Change the field \""+memoryLocationKey.toNameStringWithLocation()+"\" to volatile.");
				/*
				message.add("Change the field in line "+
						lastInstruction.getFileLocation().trim()  +
						": \""+lastInstruction.getSourceLine().trim()+"\" to volatile");
						*/
		}  

		// compute HBEdges
		Range target = new Range(raceSourceI.index,raceSourceI.instr,raceSourceI.threadId,index2,lastInstruction,threadIndex);
		ArrayList<Range> hbEdges = findHBEdges(path,memoryObjects);
		ArrayList<Integer> reported = new ArrayList<Integer>(); 
		
		// 
		// SUGGESTION (2) move source instruction
		//
		for(int i=0; i<hbEdges.size(); i++)
		{
			if (target.intersects(hbEdges.get(i)) 
					&& hbEdges.get(i).before(target) 
					&& target.lowTid == hbEdges.get(i).lowTid  
                                        && target.highTid == hbEdges.get(i).highTid
					&& !reported.contains(hbEdges.get(i).low)
					&& (!hbEdges.get(i).lowIns.getMethodInfo().getClassInfo().isSystemClass()) // By KyungHee
				/* && in the same basic block HOW TO CHECK?????   */) 
			{
			        boolean suggestionSafe = true;
			
			        /* Check if any access x between hbEdgest.get[i].low and raceSourceI.index */
			       
			        ArrayList<Integer> accessList = new ArrayList<Integer>(); 
			 
			        for(int i2=sourceInstructionNo-1; i2> hbEdges.get(i).low; i2--)
			        {	// By KyungHee    	
			           if (memoryObjects.get(i2)!=null && memoryObjects.get(i2).equals(memoryObjects.get(sourceInstructionNo)))
			              accessList.add(i2);
			        }     
			        for(int i2=0; i2<accessList.size(); i2++)
			        {  
			           int accessIndex = accessList.get(i2);
			           InstructionInfo accessInstr = getInstruction(accessIndex,path);
			           int lowIndex = hbEdges.get(i).low+1;
			           InstructionInfo lowInstr = getInstruction(lowIndex,path);
			           int highIndex = accessIndex-1;
			           InstructionInfo highInstr = getInstruction(highIndex,path);
			           Range tempRange = new Range(lowIndex,lowInstr.instr,lowInstr.threadId,highIndex,highInstr.instr,highInstr.threadId);
			           boolean hbfound = false;
			           for(int i3=i; i3 <= hbEdges.size(); i3++)
			           {
			              /* hbedges are sorted based on high, in descending order */
			              if (tempRange.low >= hbEdges.get(i3).high)
			                 break; 
			              
			              if (tempRange.includes(hbEdges.get(i3)))
			              {
			                  Range candidateEdge = hbEdges.get(i3); 
			                  /* Check for happen-before transitivity: raceSourceI.threadId == candidateEdge.lowTid AND candidateEdge.highTid == thread id of instruction accessIndex */ 
			                  if (raceSourceI.threadId == candidateEdge.lowTid && candidateEdge.highTid == accessInstr.threadId)
			                  {
			                      hbfound = true;
			                      break;
			                  }
			              }
			              
			           }  
			           if (!hbfound)
			           {
			              suggestionSafe = false;
			              break;
			           }     
                                }

			        if (!suggestionSafe) continue;
			        
			        /* Find the first access x that comes before hbEdgest.get[i].low */
			        int accessIndex = -1;
		                for(int i2=hbEdges.get(i).low-1; i2>= 0; i2--)
		                	// By KyungHee
                                   if (memoryObjects.get(i2)!=null && memoryObjects.get(i2).equals(memoryObjects.get(sourceInstructionNo)))
                                   { 
			              accessIndex = i2;
			              break;
			           }                                                                        
			        
			        if (accessIndex != -1)
			        {     
			            InstructionInfo accessInstr = getInstruction(accessIndex,path);
			            int lowIndex = accessIndex+1;
			            InstructionInfo lowInstr = getInstruction(lowIndex,path);
			            int highIndex = hbEdges.get(i).low-1;
			            InstructionInfo highInstr = getInstruction(highIndex,path);
			            Range tempRange = new Range(lowIndex,lowInstr.instr,lowInstr.threadId,highIndex,highInstr.instr,highInstr.threadId);
			            boolean hbfound = false; 
			            for(int i3=i; i3 <= hbEdges.size(); i3++)
			            {
			               /* hbedges are sorted based on high, in descending order */
			               if (tempRange.low >= hbEdges.get(i3).high)
			                  break;
			              if (tempRange.includes(hbEdges.get(i3)))
			              {
			                 Range candidateEdge = hbEdges.get(i3);
			                 /* Check for happen-before transitivity: raceSourceI.threadId == candidateEdge.lowTid AND candidateEdge.highTid == thread id of instruction at accessIndex */
			                 if (accessInstr.threadId == candidateEdge.lowTid && candidateEdge.highTid == raceSourceI.threadId)
			                 {
			                    hbfound = true;
			                    break;
			                 }
			              }    
			            }
			            if (!hbfound)
			            {
			               suggestionSafe = false;
			               break;
			            }   
			        }
			            
			                 
			        if (!suggestionSafe) continue;
			        
				message.add("Move \""+raceSourceI.instr.getSourceLine().trim()+ "\" at " + raceSourceI.instr.getFileLocation().trim() + 
						" before \"" + hbEdges.get(i).lowIns.getSourceLine().trim() +  "\" at " 
						+ hbEdges.get(i).lowIns.getFileLocation().trim()+"\n");
                                reported.add(hbEdges.get(i).low);						                          
			}                      
		}
		
		// 
		// SUGGESTION (3) synchronize on released memory location
		//
		ArrayList<KeyLine> syncLoc = findMemoryLocationsToSynchronizeOn(path, raceSourceI, stepCounter, memoryObjects);
		for(KeyLine kl : syncLoc)
			message.add("Put \""+errorInstr.getSourceLine().trim() + "\" at " + 
					errorInstr.getFileLocation().trim() +
					" inside synchronized block\n\t using the object \""+kl.key.toNameString()+ "\" of "+ kl.sourceLine); 

		// 
		// SUGGESTION (4) change other field to volatile
		//
		ArrayList<Key> volLoc = findMemoryLocationsToChangeVolatile(path, raceSourceI, stepCounter, memoryObjects, target);
		for(Key k : volLoc)
			if ( k.isField() )
			{
				message.add("Change the field \""+k.toNameStringWithLocation()+"\" to volatile.");
			}
			else
			{
				message.add("Change the type of the array field \""+k.toNameStringWithLocation()+"\" to use java.util.concurrent.atomic package.");
			}
		return message; // +timeMSG;
	}
	

	
	// This routine is copied from JPF publisher
	public static String getCounterexamplePath(JVM jvm, Path path, int stepCount)
	{
		String counterexamplePath=new String("");
		int i=0;
		int step=0;
		 // the various degrees of information for program traces
		  boolean  showCG = jvm.getConfig().getBoolean("report.console.show_cg", true);
		  boolean  showSteps = jvm.getConfig().getBoolean("report.console.show_steps", true);
		  boolean  showLocation = jvm.getConfig().getBoolean("report.console.show_location", true);
		  boolean  showSource = jvm.getConfig().getBoolean("report.console.show_source", true);
		  boolean  showMethod = jvm.getConfig().getBoolean("report.console.show_method", false);
		  boolean  showCode = jvm.getConfig().getBoolean("report.console.show_code", false);

		if (path.size() == 0) 
		      return counterexamplePath; // nothing to publish
		   
		for (Transition t : path) {
	        counterexamplePath += ("------------------------------------------------------ ");
	        counterexamplePath +=("transition #" + i++ + " thread: " + t.getThreadIndex()+"\n");

	        if (showCG){
	          counterexamplePath +=(t.getChoiceGenerator()+"\n");
	        }

	        if (showSteps) {
	          String lastLine = null;
	          MethodInfo lastMi = null;
	          int nNoSrc=0;

	          for (Step s : t) {
	        	++step;
	        	if ( step > stepCount )
	        		return counterexamplePath;
	            if (showSource) {
	              String line = s.getLineString();
	              if (line != null) {
	                if (!line.equals(lastLine)) {
	                  if (nNoSrc > 0){
	                    counterexamplePath +=("      [" + nNoSrc + " insn w/o sources]"+"\n");
	                  }

	                  counterexamplePath += ("  ");
	                  if (showLocation) {
	                    counterexamplePath += (Left.format(s.getLocationString(),30));
	                    counterexamplePath += (" : ");
	                  }
	                  counterexamplePath +=(line.trim()+"\n");
	                  nNoSrc = 0;

	                }
	              } else { // no source
	                nNoSrc++;
	              }
	              
	              lastLine = line;
	            }

	            if (showCode) {
	              Instruction insn = s.getInstruction();
	              if (showMethod){
	                MethodInfo mi = insn.getMethodInfo();
	                if (mi != lastMi) {
	                  ClassInfo mci = mi.getClassInfo();
	                  counterexamplePath += ("    ");
	                  if (mci != null) {
	                    counterexamplePath += (mci.getName());
	                    counterexamplePath += (".");
	                  }
	                  counterexamplePath +=(mi.getUniqueName()+"\n");
	                  lastMi = mi;
	                }
	              }
	              counterexamplePath += ("      ");
	              counterexamplePath +=(insn+"\n");
	            }
	          }

	          if (showSource && !showCode && (nNoSrc > 0)) {
	            counterexamplePath +=("      [" + nNoSrc + " insn w/o sources]"+"\n");
	          }
	        }
	      }
		return counterexamplePath;
	}
	
	public static String getSnapshot()
	{
	    java.io.StringWriter sw = new java.io.StringWriter();
	    java.io.PrintWriter pw = new java.io.PrintWriter(sw);
	    JVM.getVM().printLiveThreadStatus(pw);
		return ""+sw;
	}
}
